@@ -1,11 +1,16 @@
 """Encapsulate the lookup of packages and their provided import names."""
 
 import logging
+import subprocess
 import sys
+import tempfile
+import venv
+from abc import ABC, abstractmethod
+from contextlib import contextmanager
 from dataclasses import dataclass, field
 from enum import Enum
 from pathlib import Path
-from typing import Dict, Iterable, Optional, Set
+from typing import Dict, Iterable, Iterator, List, Optional, Set
 
 # importlib_metadata is gradually graduating into the importlib.metadata stdlib
 # module, however we rely on internal functions and recent (and upcoming)
@@ -76,32 +81,29 @@ class Package:
         self.mappings.setdefault(mapping, set()).update(import_names)
         self.import_names.update(import_names)
 
-    def add_identity_import(self) -> None:
-        """Add identity mapping to this package.
-
-        This builds on an assumption that a package 'foo' installed with e.g.
-        `pip install foo`, will also provide an import name 'foo'. This
-        assumption does not always hold, but sometimes we don't have much else
-        to go on...
-        """
-        self.add_import_names(
-            self.normalize_name(self.package_name),
-            mapping=DependenciesMapping.IDENTITY,
-        )
-
-    @classmethod
-    def identity_mapping(cls, package_name: str) -> "Package":
-        """Factory for conveniently creating identity-mapped package object."""
-        ret = cls(package_name)
-        ret.add_identity_import()
-        return ret
-
     def is_used(self, imported_names: Iterable[str]) -> bool:
         """Return True iff this package is among the given import names."""
         return bool(self.import_names.intersection(imported_names))
 
 
-class LocalPackageLookup:
+class BasePackageResolver(ABC):
+    """Define the interface for doing package -> import names lookup."""
+
+    @abstractmethod
+    def lookup_packages(self, package_names: Set[str]) -> Dict[str, Package]:
+        """Convert package names into a Package objects with available imports.
+
+        Resolve as many of the given package names as possible into their
+        corresponding import names, and return a dict that maps the resolved
+        names to their corresponding Package objects.
+
+        Return an empty dict if this PackageResolver is unable to resolve any
+        of the given packages.
+        """
+        raise NotImplementedError
+
+
+class LocalPackageResolver(BasePackageResolver):
     """Lookup imports exposed by packages installed in a Python environment."""
 
     def __init__(self, pyenv_path: Optional[Path] = None) -> None:
@@ -176,50 +178,131 @@ class LocalPackageLookup:
 
         return self._packages
 
-    def lookup_package(self, package_name: str) -> Optional[Package]:
-        """Convert a package name to a locally available Package object.
+    def lookup_packages(self, package_names: Set[str]) -> Dict[str, Package]:
+        """Convert package names to locally available Package objects.
 
         (Although this function generally works with _all_ locally available
         packages, we apply it only to the subset that is the dependencies of
         the current project.)
 
-        Return the Package object that encapsulates the package-name-to-import-
-        names mapping for the given package name.
+        Return a dict mapping package names to the Package objects that
+        encapsulate the package-name-to-import-names mappings.
 
-        Return None if we're unable to find any import names for the given
-        package name. This is typically because the package is missing from the
-        current environment, or because we fail to determine its provided import
-        names.
+        Only return dict entries for the packages that we manage to find in the
+        local environment. Omit any packages for which we're unable to determine
+        what imports names they provide. This applies to packages that are
+        missing from the local environment, or packages where we fail to
+        determine its provided import names.
         """
-        return self.packages.get(Package.normalize_name(package_name))
+        return {
+            name: self.packages[Package.normalize_name(name)]
+            for name in package_names
+            if Package.normalize_name(name) in self.packages
+        }
+
+
+class TemporaryPipInstallResolver(BasePackageResolver):
+    """Resolve packages by installing them in to a temporary venv.
+
+    This provides a resolver for packages that are not installed in an existing
+    local environment. This is done by creating a temporary venv, and then
+    `pip install`ing the packages into this venv, and then resolving the
+    packages in this venv. The venv is automatically deleted before as soon as
+    the packages have been resolved."""
+
+    @staticmethod
+    @contextmanager
+    def temp_installed_requirements(requirements: List[str]) -> Iterator[Path]:
+        """Create a temporary venv and install the given requirements into it.
+
+        Provide a path to the temporary venv into the caller's context in which
+        the given requirements have been `pip install`ed. Automatically remove
+        the venv at the end of the context.
+        """
+        with tempfile.TemporaryDirectory() as tmpdir:
+            venv_dir = Path(tmpdir)
+            venv.create(venv_dir, clear=True, with_pip=True)
+            subprocess.run(
+                [f"{venv_dir}/bin/pip", "install", "--no-deps"] + requirements,
+                check=True,  # fail if any of the commands fail
+            )
+            (venv_dir / ".installed").touch()
+            yield venv_dir
+
+    def lookup_packages(self, package_names: Set[str]) -> Dict[str, Package]:
+        """Convert package names into Package objects via temporary pip install.
+
+        Use the temp_installed_requirements() above to `pip install` the given
+        package names into a temporary venv, and then use LocalPackageResolver
+        on this venv to provide the Package objects that correspond to the
+        package names.
+        """
+        logger.info("Installing dependencies into a new temporary Python environment.")
+        with self.temp_installed_requirements(sorted(package_names)) as venv_dir:
+            local_resolver = LocalPackageResolver(venv_dir)
+            return local_resolver.lookup_packages(package_names)
+
+
+class IdentityMapping(BasePackageResolver):
+    """An imperfect package resolver that assumes package name == import name.
+
+    This will resolve _any_ package name into a corresponding identical import
+    name (modulo normalization, see Package.normalize_name() for details).
+    """
+
+    @staticmethod
+    def lookup_package(package_name: str) -> Package:
+        """Convert a package name into a Package with the same import name."""
+        ret = Package(package_name)
+        import_name = Package.normalize_name(package_name)
+        ret.add_import_names(import_name, mapping=DependenciesMapping.IDENTITY)
+        logger.info(
+            f"{package_name!r} was not resolved. "
+            f"Assuming it can be imported as {import_name!r}."
+        )
+        return ret
+
+    def lookup_packages(self, package_names: Set[str]) -> Dict[str, Package]:
+        """Convert package names into Package objects w/the same import name."""
+        return {name: self.lookup_package(name) for name in package_names}
 
 
 def resolve_dependencies(
-    dep_names: Iterable[str], pyenv_path: Optional[Path] = None
+    dep_names: Iterable[str],
+    pyenv_path: Optional[Path] = None,
+    install_deps: bool = False,
 ) -> Dict[str, Package]:
     """Associate dependencies with corresponding Package objects.
 
-    Use LocalPackageLookup to find Package objects for each of the given
+    Use LocalPackageResolver to find Package objects for each of the given
     dependencies inside the virtualenv given by 'pyenv_path'. When 'pyenv_path'
     is None (the default), look for packages in the current Python environment
     (i.e. equivalent to sys.path).
 
-    For dependencies that cannot be found with LocalPackageLookup,
-    fabricate an identity mapping (a pseudo-package making available an import
-    of the same name as the package, modulo normalization).
-
     Return a dict mapping dependency names to the resolved Package objects.
     """
-    ret = {}
-    local_packages = LocalPackageLookup(pyenv_path)
-    for name in dep_names:
-        if name not in ret:
-            package = local_packages.lookup_package(name)
-            if package is None:  # fall back to identity mapping
-                package = Package.identity_mapping(name)
-                logger.info(
-                    f"Could not find {name!r} in the current environment. Assuming "
-                    f"it can be imported as {', '.join(sorted(package.import_names))}"
-                )
-            ret[name] = package
+    deps = set(dep_names)  # consume the iterable once
+
+    # This defines the "stack" of resolvers that we will use to convert
+    # dependencies into provided import names. We call .lookup_package() on
+    # each resolver in order until one of them returns a Package object. At
+    # that point we are happy, and don't consult any of the later resolvers.
+    resolvers: List[BasePackageResolver] = [LocalPackageResolver(pyenv_path)]
+    if install_deps:
+        resolvers += [TemporaryPipInstallResolver()]
+    # Identity mapping being at the bottom of the resolvers stack ensures that
+    # _all_ deps are matched. TODO: If we make the identity mapping optional,
+    # we must remember to properly handle/signal unresolved dependencies.
+    resolvers += [IdentityMapping()]
+
+    ret: Dict[str, Package] = {}
+    for resolver in resolvers:
+        unresolved = deps - ret.keys()
+        if not unresolved:  # no unresolved deps left
+            logger.debug("No dependencies left to resolve!")
+            break
+        logger.debug(f"Trying to resolve {unresolved!r} with {resolver}")
+        resolved = resolver.lookup_packages(unresolved)
+        logger.debug(f"  Resolved {resolved!r} with {resolver}")
+        ret.update(resolved)
     return ret
